/***********************************************************************
 * HT1632_LedMatrix.cpp - Arduino library for Holtek HT1632 LED driver chip,
 *     As implemented on the Sure Electronics DE-DP10X display board
 *       (8 x 32 dot matrix LED module.)
 *
 * Original code by:
 * Nov, 2008 by Bill Westfield ("WestfW")
 *   Copyrighted and distributed under the terms of the Berkely license
 *   (copy freely, but include this notice of original author.)
 *
 * Adapted for 8x32 display by FlorinC.
 *
 * Library Created and updated by Andrew Lindsay October/November 2009
 ***********************************************************************/

#include "HT1632_LedMatrix.h"
#include <inttypes.h>
#include <wprogram.h>
#include <avr/pgmspace.h>
#include "font3.h"

// Display size and configuration, defaul is for a single 8x32 display
byte numDevices = 0;	// Total number of devices
byte numXDevices = 1;	// Number of horizontal devices
byte numYDevices = 1;	// Number of vertical devices
byte xMax = 32 * numXDevices - 1;
byte xSize = 32 * numXDevices;
byte yMax = 8 * numYDevices - 1;
byte ySize = 8 * numYDevices;

/*
 * Set these constants to the values of the pins connected to the SureElectronics Module
 * NOTE - these are different from the original demo code by westfw
 *
 * Use Pin Mappings 8-11 for CS1 to 4, 12 for Data and 13 for Clock
 * Use mixture of #define and variables.
 * Pin numbers are for setup and port identifiers are for direct port writes.
 */

#define LOWPINS

#ifdef LOWPINS
// Port D defines for 2-7
// If using Nuelectronics ethernet shield, need to cut unused interrupt track to PD2
byte ht1632_cs[4] = {2, 3, 4, 5};    // Chip Select (1, 2, 3, or 4)
volatile uint8_t* ht1632_cs_Port[4] = {&PORTD, &PORTD, &PORTD, &PORTD};    // Chip Select (1, 2, 3, or 4)
volatile uint8_t* ht1632_cs_DDR[4] = {&DDRD, &DDRD, &DDRD, &DDRD};    // Chip Select (1, 2, 3, or 4)
#define HT1632_DATA       6      // Data pin
#define HT1632_DATA_PORT  PORTD  // Data port
#define HT1632_DATA_DDR   DDRD   // Data port
#define HT1632_WRCLK      7      // Write clock pin
#define HT1632_WRCLK_PORT PORTD  // Write clock port
#define HT1632_WRCLK_DDR  DDRD   // Write clock port

#else
// Port B defines for 8-13
byte ht1632_cs[4] = {8, 9, 10, 11};    // Chip Select (1, 2, 3, or 4)
volatile uint8_t* ht1632_cs_Port[4] = {&PORTB, &PORTB, &PORTB, &PORTB};    // Chip Select (1, 2, 3, or 4)
volatile uint8_t* ht1632_cs_DDR[4] = {&DDRB, &DDRB, &DDRB, &DDRB};    // Chip Select (1, 2, 3, or 4)
#define HT1632_DATA     12       // Data pin
#define HT1632_DATA_PORT  PORTB  // Data port
#define HT1632_DATA_DDR   DDRB   // Data port
#define HT1632_WRCLK    13       // Write clock pin
#define HT1632_WRCLK_PORT PORTB  // Write clock port
#define HT1632_WRCLK_DDR  DDRB   // Write clock port

#endif

// helper macros
#define output_low(port,pin) port &= ~(1<<pin)
#define output_high(port,pin) port |= (1<<pin)

// Custom character buffers - 8 characters available
byte cgram [8][8] = { 
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 } };

// Variables used to keep track of cursor position
int cursorX = 0;
int cursorY = 0;

// Default constructor
HT1632_LedMatrix::HT1632_LedMatrix( void ) {
}

void HT1632_LedMatrix::init( byte xDevices, byte yDevices ) {
  // Set up the display size based on number of devices both horizontal and vertical
  numXDevices = xDevices;
  xMax = 32 * numXDevices - 1;
  xSize = 32 * numXDevices;
  numYDevices = yDevices;
  yMax = 8 * numYDevices - 1;
  ySize = 8 * numYDevices;

  numDevices = numXDevices * numYDevices;

  HT1632_WRCLK_DDR |= _BV( HT1632_WRCLK );		// set pins to output
  HT1632_DATA_DDR |= _BV( HT1632_DATA );

  for (byte chipno=0; chipno<numDevices; chipno++){
	  *ht1632_cs_DDR[chipno] |= _BV( ht1632_cs[chipno] );	// output pin
	  chipfree(chipno); 	// unselect it 
	  sendcmd(chipno, HT1632_CMD_SYSDIS);  // Disable system
	  sendcmd(chipno, HT1632_CMD_COMS00);  // 08*32, PMOS drivers
	  sendcmd(chipno, HT1632_CMD_RCCLK); 	// Master Mode 
	  sendcmd(chipno, HT1632_CMD_SYSON); 	// System on 
	  sendcmd(chipno, HT1632_CMD_LEDON); 	// LEDs on 
	  sendcmd(chipno, HT1632_CMD_PWM | 0x0c); 	// PWM Duty 
	  for (byte i=0; i<96; i++)
		senddata(chipno, i, 0);  // clear the display
	  delay(100);
  }
  cursorX = 0;
  cursorY = 0;
}
 
/***********************************************************************
 * chipselect / chipfree
 * Select or de-select a particular ht1632 chip.
 * De-selecting a chip ends the commands being sent to a chip.
 * CD pins are active-low; writing 0 to the pin selects the chip.
 ***********************************************************************/
void HT1632_LedMatrix::chipselect(byte chipno)
{
  output_low(*ht1632_cs_Port[chipno], ht1632_cs[chipno]);
}

void HT1632_LedMatrix::chipfree(byte chipno)
{
  output_high(*ht1632_cs_Port[chipno], ht1632_cs[chipno]);
}

/*
 * we keep a copy of the display controller contents so that we can
 * know which bits are on without having to (slowly) read the device.
 * Note that we only use the low four bits of the shadow ram, since
 * we're shadowing 4-bit memory.  This makes things faster, and we
 * use the other half for a "snapshot" when we want to plot new data
 * based on older data...
 */
// (fc) covers the case for 32x8 as well (64 bytes, 4 bits)
byte shadowram[128];  // our copy of the display's RAM

/*
 * writebits
 * Write bits to h1632 on pins HT1632_DATA, HT1632_WRCLK
 * Chip is assumed to already be chip-selected
 * Bits are shifted out from MSB to LSB, with the first bit sent
 * being (bits & firstbit), shifted till firsbit is zero.
 */
void HT1632_LedMatrix::writebits (byte bits, byte firstbit)
{
  while (firstbit) {
    output_low(HT1632_WRCLK_PORT, HT1632_WRCLK);
    if (bits & firstbit) {
      output_high(HT1632_DATA_PORT,HT1632_DATA);
    } else {
      output_low(HT1632_DATA_PORT, HT1632_DATA);
    }
    output_high(HT1632_WRCLK_PORT, HT1632_WRCLK);
    firstbit >>= 1;
  }
}

/*
 * sendcmd
 * Send a command to the ht1632 chip.
 * A command consists of a 3-bit "CMD" ID, an 8bit command, and
 * one "don't care bit".
 *   Select 1 0 0 c7 c6 c5 c4 c3 c2 c1 c0 xx Free
 */
void HT1632_LedMatrix::sendcmd (byte chipno, byte command)
{
  chipselect(chipno);  // Select chip
  writebits(HT1632_ID_CMD, 0x04);	//1<<2);  // send 3 bits of id: COMMMAND
  writebits(command, 0x80);	//1<<7);  // send the actual command
  writebits(0, 1); 	/* one extra dont-care bit in commands. */
  chipfree(chipno); //done
}

/*
 * clear
 * clear the display, and the shadow memory, and the snapshot
 * memory.  This uses the "write multiple words" capability of
 * the chipset by writing all 96 words of memory without raising
 * the chipselect signal.
 */
void HT1632_LedMatrix::clear()
{
  char i;

  for (byte chipno=0; chipno<numDevices; chipno++){
    chipselect(chipno);  // Select chip
    writebits(HT1632_ID_WR, 0x04);	//1<<2);  // send ID: WRITE to RAM
    writebits(0, 0x40);		//1<<6); // Send address
    for (i = 0; i < 32; i++) // Clear entire display
      writebits(0, 0x80);	//1<<7); // send 8 bits of data
    chipfree(chipno); // done
    for (i=0; i < 64; i++)
      shadowram[i+64*chipno] = 0;
  }
  cursorX = 0;
  cursorY = 0;
}


// Brighness is from 0 to 15
void HT1632_LedMatrix::setBrightness( unsigned char brightness ) {
  for (byte chipno=0; chipno<numDevices; chipno++) {
	sendcmd(chipno, HT1632_CMD_PWM | (brightness & 0x0F ));
  }
}


/*
 * senddata
 * send a nibble (4 bits) of data to a particular memory location of the
 * ht1632.  The command has 3 bit ID, 7 bits of address, and 4 bits of data.
 *    Select 1 0 1 A6 A5 A4 A3 A2 A1 A0 D0 D1 D2 D3 Free
 * Note that the address is sent MSB first, while the data is sent LSB first!
 * This means that somewhere a bit reversal will have to be done to get
 * zero-based addressing of words and dots within words.
 */
void HT1632_LedMatrix::senddata (byte chipno, byte address, byte data)
{
  chipselect(chipno);  // Select chip
  writebits(HT1632_ID_WR, 0x04);	//1<<2);  // send ID: WRITE to RAM
  writebits(address, 0x40);		//1<<6); // Send address
  writebits(data, 0x08);		//1<<3); // send 4 bits of data
  chipfree(chipno); // done
}

// Write a string at the position specified
void HT1632_LedMatrix::putString(int x, int y, char *str) {
	cursorX = x;
	cursorY = y;
	while( *str ) {
		putChar( cursorX, y, *str++ );
		cursorX += 6;
	}
}

/*
 * Copy a character glyph from the myfont data structure to
 * display memory, with its upper left at the given coordinate
 * This is unoptimized and simply uses plot() to draw each dot.
 */
void HT1632_LedMatrix::write( uint8_t c) {
	putChar( cursorX, cursorY, (char)c );
}

/*
 * Copy a character glyph from the myfont data structure to
 * display memory, with its upper left at the given coordinate
 * This is unoptimized and simply uses plot() to draw each dot.
 */
void HT1632_LedMatrix::putChar(int x, int y, char c) {
  // fonts defined for ascii 32 and beyond (index 0 in font array is ascii 32);
  // CGRAM characters are in range 0 to 15 with 8-15 being repeat of 0-7
  byte charIndex;
  cursorX = x;
  cursorY = y;

  if( c > 15 ) {
	// Regular characters
  	// replace undisplayable characters with blank;
  	if (c < 32 || c > 126) {
    		charIndex = 0;
  	} else {
    		charIndex = c - 32;
  	}

  	// move character definition, pixel by pixel, onto the display;
  	// fonts are defined as one byte per row;
  	for (byte row=0; row<8; row++) {
    		byte rowDots = pgm_read_byte_near(&myfont[charIndex][row]);
    		for (byte col=0; col<6; col++) {
      			if (rowDots & (1<<(5-col)))
        			plot(x+col, y+row, 1);
      			else
        			plot(x+col, y+row, 0);
    		}
  	}
  } else {
	// CGRAM Characters
	charIndex = c & 0x07;		// Only low 3 bits count
  	// move character definition, pixel by pixel, onto the display;
  	// fonts are defined as one byte per row;
  	for (byte row=0; row<8; row++) {
    		byte rowDots = cgram[charIndex][row];
    		for (byte col=0; col<6; col++) {
      			if (rowDots & (1<<(5-col)))
        			plot(x+col, y+row, 1);
      			else
        			plot(x+col, y+row, 0);
    		}
  	}

  }

  cursorX += 6;
}

// Set position of cursor for writing
void HT1632_LedMatrix::gotoXY(int x, int y) {
	cursorX = x;
	cursorY = y;
}

// Shift cursor X position a number of positions either left or right.
void HT1632_LedMatrix::shiftCursorX(int xinc) {
	cursorX += xinc;
}


/*
 * plot a point on the display, with the upper left hand corner
 * being (0,0), and the lower right hand corner being (23, 15).
 * Note that Y increases going "downward" in contrast with most
 * mathematical coordiate systems, but in common with many displays
 * No error checking; bad things may happen if arguments are out of
 * bounds!  (The ASSERTS compile to nothing by default
 */
void HT1632_LedMatrix::plot (int x, int y, char val)
{
  if (x<0 || x>xMax || y<0 || y>yMax)
     return;

  char addr, bitval;
  byte chipno = (x >> 5) + (y >> 3);

  /*
   * The 4 bits in a single memory word go DOWN, with the LSB
   * (first transmitted) bit being on top.  However, writebits()
   * sends the MSB first, so we have to do a sort of bit-reversal
   * somewhere.  Here, this is done by shifting the single bit in
   * the opposite direction from what you might expect.
   */
  bitval = 8>>(y&3);  // compute which bit will need set

  addr = ((x%32)<<1) + (y>>2);  // compute which memory word this is in

  if (val) {  // Modify the shadow memory
    shadowram[addr+64*chipno] |= bitval;
  }
  else {
    shadowram[addr+64*chipno] &= ~bitval;
  }
  // Now copy the new memory value to the display
//  senddata(qht1632_cs[chipno], addr, shadowram[addr+64*chipno]);
  senddata(chipno, addr, shadowram[addr+64*chipno]);
}


void HT1632_LedMatrix::setCustomChar( int charNum, unsigned char cgchar[] ) {
	for(int i=0; i<8; i++ ) {
		cgram[charNum][i] = (byte)cgchar[i];
	}
}



// The end!
